module Sol_Exercise_11 where

import Control.Monad
import Data.Time
import Network
import System.IO
import System.Locale

import Data.List as List

import qualified Sol_AssocList as AL
import qualified Sol_Map2 as M2


{- Library -- nicht veraendern -}

vocabs = ["i", "more", "more", "now", "want", "won", "wow"]

getTime :: String -> IO String
getTime str = do
  time <- getZonedTime
  let output = formatTime defaultTimeLocale str time
  return output

{- G11.1 -}

{- Siehe Sol_AssocList.hs -}

{- G11.2 -}

{- Siehe Sol_Queue.hs -}

{- G11.3 -}

quasiIdentical :: String -> String -> Bool
quasiIdentical [] [] = False
quasiIdentical (c : cs) (d : ds)
    | c == d = quasiIdentical cs ds
    | otherwise = cs == ds
quasiIdentical _ _ = False

safeFixTypo :: [String] -> String -> Either (String, [String]) String
safeFixTypo vocabs word =
    if word `elem` vocabs then Right word else Left (word, quasis)
  where quasis = List.nub (List.filter (quasiIdentical word) vocabs)


safeFixTypos :: [String] -> [String] -> [Either (String, [String]) String]
safeFixTypos = map . safeFixTypo

{- H11.1 -}

{- Siehe Sol_Map2.hs -}

assocListToMap2 :: Ord k => AL.Map k v -> M2.Map k v
assocListToMap2 al = foldl' ins M2.empty (AL.keys al)
  where
    ins m2 k = case AL.lookup k al of
        Nothing -> m2
        Just v -> M2.insert k v m2


map2ToAssocList :: Ord k => M2.Map k v -> AL.Map k v
map2ToAssocList m2 = foldl' ins AL.empty (M2.keys m2)
  where
    ins al k = case M2.lookup k m2 of
        Nothing -> al
        Just v -> AL.insert k v al

{-
 - Um den Inhalt einer Map aus dem Modul AL (oder M2) zu erhalten,
 - können wir mit AL.keys die Liste der Schlüssel und
 - dann mit AL.lookup zu jedem Schlüssel den Wert bestimmen.
 -
 - Dabei wird (im Fall von AL) die Liste mehrfach durchlaufen (was zu einem
 - quadratischen Aufwand führt). Ähnlich ist das mit unserer Implementierung
 - von M2.
 -
 - Außerdem müssen wir in den Konvertierungsfunktionen noch den Fall behandeln,
 - dass der Lookup Nothing zurückliefert, obwohl dies nicht auftreten kann.
 -
 - Eine Lösungsansatz wäre es, die Module jeweils um eine Funktion zu erweitern,
 - die eine Liste der in der Map enthaltenen Schlüssel-Wert-Paare zurückgeben:
 -
 -  AL:     kvList ::  Eq k => Map k v -> [(k,v)]
 -  M2:     kvList :: Ord k => Map k v -> [(k,v)]
 -
 - Die offensichtliche Implementierung in AL ist natürlich
 -
 -    kvList (AL al) = al
 -
 - Dennoch wird hier das Prinzip des abstrakten Datentyps gewahrt: Die interne
 - Repräsentation ist nur zufällig mit spezifizierten Ausgabe von kvList identisch.
 -
 - Damit sind effiziente (und kompakte) Implementierungen möglich:
 -}

assocListToMap2' :: Ord k => AL.Map k v -> M2.Map k v
assocListToMap2' = foldl' (\m2 (k,v) -> M2.insert k v m2) M2.empty . AL.kvList

map2ToAssocList' :: Ord k => M2.Map k v -> AL.Map k v
map2ToAssocList' = foldl' (\al (k,v) -> AL.insert k v al) AL.empty . M2.kvList

{-
 - Unter anderem um zu vermeiden, dass eine zusätzliche Liste erzeugt wird,
 - gibt man häufig statt kvList direkt eine fold-Funktion auf der Map an:
 -
 - AL:      fold ::  Eq k => (k -> v -> a -> a) -> a -> Map k v -> a
 - M2:      fold :: Ord k => (k -> v -> a -> a) -> a -> Map k v -> a
 -}
assocListToMap2'' :: Ord k => AL.Map k v -> M2.Map k v
assocListToMap2'' = AL.fold M2.insert M2.empty

map2ToAssocList'' :: Ord k => M2.Map k v -> AL.Map k v
map2ToAssocList'' = M2.fold AL.insert AL.empty


{- H11.2 -}

serverLoop :: Socket -> IO ()
serverLoop sock = forever loop
  where loop = do (h, _, _) <- accept sock
                  hPutStrLn h "I2B11H2P/1.0 HELLO"
                  request <- hGetLine h
                  response <- time $ words request
                  hPutStrLn h response
                  hClose h
        time ["DATE"] = getTime "%Y-%m-%d"
        time _        = getTime "%H:%M:%S"

server :: IO ()
server = withSocketsDo $
  listenOn (PortNumber 9002) >>= serverLoop

{- H11.3 -}

{-WETT-}
{-TTEW-}
